{
  open RubyParser
  open RubyLexerState

  let remove_underscores str = 
    let len = String.length str in
    let buf = Buffer.create len in
      for i = 0 to (len-1); do
	match str.[i] with
	  | '.'
	  | '-' (* needed for 2.3e-4 *)
	  | 'e'
	  | ('a'..'f') | ('A'..'F') | 'x'
	  | ('0'..'9') as i -> Buffer.add_char buf i
	  | '_' -> ()
	  | _ -> failwith "non number and non-'_' in fixnum"
      done;
      Buffer.contents buf

  let to_bignum str = Big_int.big_int_of_string (remove_underscores str)

  let incr_line lexbuf =
    let pos = lexbuf.Lexing.lex_curr_p in
    lexbuf.Lexing.lex_curr_p <- { pos with
      Lexing.pos_lnum = pos.Lexing.pos_lnum + 1;
      Lexing.pos_bol = pos.Lexing.pos_cnum;
    }

  let eol_or_heredoc f_lead f_no_lead state lexbuf =
    if Queue.is_empty state.heredoc_queue
    then T_EOL
    else let (lead,i,d) = Queue.pop state.heredoc_queue in 
    let buf = Buffer.create 32 in
      if lead then f_lead state buf i d lexbuf 
      else f_no_lead state buf i d lexbuf

  let num_of_hex_digit = function
    | '0' -> 0
    | '1' -> 1
    | '2' -> 2
    | '3' -> 3
    | '4' -> 4
    | '5' -> 5
    | '6' -> 6
    | '7' -> 7
    | '8' -> 8
    | '9' -> 9
    | 'a' | 'A' -> 10
    | 'b' | 'B' -> 11
    | 'c' | 'C' -> 12
    | 'd' | 'D' -> 13
    | 'e' | 'E' -> 14
    | 'f' | 'F' -> 15
    | _ -> failwith "num_of_hex_digit"

  let big_int_of_hex_string str = 
    if String.sub str 0 2 <> "0x" then failwith "big_int_of_hex_string";
    let base10 x exp = (* compute x * (16 ** exp) *)
      Big_int.mult_big_int x 
	(Big_int.power_int_positive_int 16 exp)
    in
    let rec helper acc idx exp = 
      if idx < 2
      then acc (* skip 0x *)
      else
	let digit = num_of_hex_digit str.[idx] in
	let b10 = base10 (Big_int.big_int_of_int digit) exp in
	let acc' = Big_int.add_big_int b10 acc in
	  helper acc' (idx - 1) (exp + 1)
    in
    let len = String.length str in
      helper Big_int.zero_big_int (len-1) 0
(*
  let predef_vars = 
  [ "$!"; "$@"; "$&"; "$`"; "$'"; "$+";
  "$1"; "$2"; "$3"; "$4"; "$5"; "$6"; "$7"; "$8"; "$9";
  "$~"; "$="; "$/"; "$\\"; "$,"; "$;"; "$."; "$<"; "$>";
      "$_"; "$0"; "$*"; "$$"; "$?"; "$:"; "$\"";
  ]
*)
}

let ws_re = ['\n' '\r' '\t' ' ']
let alpha_re = ['a'-'z''A'-'Z']
let num_re = ['0'-'9']
let alphanum_re = alpha_re | num_re
let rubynum_re = num_re ("_" | num_re)* 
let fixnum_re = rubynum_re ("e" rubynum_re)?
let rubyfloat_re = rubynum_re '.' (rubynum_re) ("e" ("-")? rubynum_re+)?
let id_start = '_' | alpha_re
let id_body = '_' | alphanum_re
let id_suffix = '?'|'!'(*|'='*)
let id_re = id_start id_body* id_suffix? 
let string_single_delim = [
  '!' '@' '#' '$' '%' '^' '&' '*'
  ',' '.' '?' '`' '~' '|' '+' '_'
  '-' '\\' '/' ':' '"' '\'']

rule token state = parse
  | '+'   {let s = state.expr_state in 
	     beg_state state; match s with
	     | Expr_Def | Expr_Beg -> T_UPLUS
	     | Expr_Mid -> T_PLUS}
  | '-'   {let s = state.expr_state in 
	     beg_state state; match s with
	     | Expr_Def | Expr_Beg -> T_UMINUS
	     | Expr_Mid -> T_MINUS}
        (*{choose_token state T_UMINUS T_MINUS}*)
  | '.'   {def_state state;T_DOT}
  | ','   {beg_state state;T_COMMA}
  | '!'   {beg_state state;T_BANG}
  | '~'   {beg_state state;T_TILDE}
  | "**"  {beg_state state;T_POW}
  | "<=>" {beg_state state;T_CMP}
(*  | " ="  {beg_state state;T_ASSIGN_OP}*)
  | "="   {beg_state state;T_ASSIGN}
  | "=="  {beg_state state;T_EQ}
  | "===" {beg_state state;T_EQQ}
  | "!="  {beg_state state;T_NEQ}
  | ">="  {beg_state state;T_GEQ}
  | "<="  {beg_state state;T_LEQ}
  | "<"   {beg_state state;T_LT}
  | ">"   {beg_state state;T_GT}
  | "&&"  {beg_state state;T_ANDOP}
  | "||"  {beg_state state;T_OROP}
  | "=~"  {beg_state state;T_MATCH}
  | "!~"  {beg_state state;T_NMATCH}
(*  | "[]="  {beg_state state;T_ASET}*)
  | (['+' '-' '*' '&' '|' '%' '^'] | "||" | "&&" | "<<" | ">>") '='
      {beg_state state; T_OP_ASGN (Lexing.lexeme lexbuf)}

  | "<<-" '\'' ([^'\'']+ as delim) '\''
  | "<<-" '"' ([^'"']+ as delim) '"'
  | "<<-" (id_body* as delim)
      {let id = fresh_heredoc state true delim in
	 mid_state state; T_HEREDOC id}
(*
  | " <<" ['\'' '"']?(id_body+ as delim)['\'' '"']?
      {let id = fresh_heredoc state false delim in
	 mid_state state; T_HEREDOC id}
    *)
  | "<<" ws_re {beg_state state; T_LSHFT}
  | "<<"
      {match state.expr_state with
	| Expr_Def -> beg_state state;T_LSHFT
	| Expr_Mid
	| Expr_Beg -> heredoc_header state lexbuf
      }

  | ">>"  {beg_state state;T_RSHFT}
  | "=>"  {beg_state state;T_ASSOC}
  | '('   {choose_token state T_LPAREN T_LPAREN_ARG}
(*  | '('   {beg_state state; T_LPAREN}*)
  | ')'   {mid_state state;T_RPAREN}
  | " ["  {beg_state state;T_LBRACK}
  | '['   {choose_token state T_LBRACK T_LBRACK_ARG}
(*  | '['   {beg_state state;T_LBRACK}*)
  | ']'   {mid_state state;T_RBRACK}
  | '{'   {beg_state state;T_LBRACE}
      (*| ',' [' ''\t''\n']* '}'   {mid_state state;T_RBRACE}*)
      (*  | ',' [' ''\t''\n']* ']'   {mid_state state;T_RBRACK}*)
  | '}'   {mid_state state;T_RBRACE}
  | '*'   {choose_token state T_USTAR T_STAR}
  | "..." {beg_state state;T_DOT3}
  | ".."  {beg_state state;T_DOT2}
  | '?'   {match state.expr_state with
	     | Expr_Mid -> beg_state state;T_QUESTION
	     | Expr_Def (* parse error anyway *)
	     | Expr_Beg -> mid_state state;char_code lexbuf
      }
  | '&'   {choose_token state T_UAMPER T_AMPER}
  | '#'   {beg_state state; comment lexbuf}
  | '`'([^'`']* as s)'`'   {T_TICK s}
  | '^'   {beg_state state;T_CARROT}
  | '|'   {beg_state state;T_VBAR}
  | ';'   {beg_state state;T_SEMICOLON}

  | '\''  {mid_state state; single_string (Buffer.create 32) lexbuf}
  | '\"'  {mid_state state; double_string (Buffer.create 32) lexbuf}

  | '%'   {
      match state.expr_state with
	| Expr_Def -> beg_state state; T_PERCENT
	| _ -> percent state lexbuf}

  | "=begin" {beg_state state; delim_comment state lexbuf}

  | fixnum_re
      { mid_state state;let num = to_bignum (Lexing.lexeme lexbuf) in
	  if Big_int.is_int_big_int num
	  then T_FIXNUM (Big_int.int_of_big_int num)
	  else T_BIGNUM num
      }
  | "0x" (num_re|['a'-'f''A'-'F''_'])+
      { mid_state state;
	let num = big_int_of_hex_string (Lexing.lexeme lexbuf) in
	  if Big_int.is_int_big_int num
	  then T_FIXNUM (Big_int.int_of_big_int num)
	  else T_BIGNUM num
      }
  | rubyfloat_re
      { mid_state state;T_FLOAT (float_of_string (remove_underscores (Lexing.lexeme lexbuf))) }

  | "\\\n" { incr_line lexbuf; token state lexbuf }
  | "\n" 
      {beg_state state; incr_line lexbuf; 
       eol_or_heredoc heredoc_lead heredoc state lexbuf
      }
  | eof   {beg_state state;T_EOF}

  | "alias"    {def_state state;K_ALIAS}
  | "and"      {beg_state state;K_AND}
  | "begin"    {beg_state state;K_lBEGIN}
  | "BEGIN"    {beg_state state;K_BEGIN}
  | "case"     {beg_state state;K_CASE}
  | "class"    {def_state state;K_CLASS}
  | "def"      {def_state state;K_DEF}
  | "defined?" {mid_state state;K_DEFINED}
  | "do"       {beg_state state;K_DO}
  | "else"     {beg_state state;K_ELSE}
  | "elsif"    {beg_state state;K_ELSIF}
  | "END"      {beg_state state;K_END}
  | "end"      {mid_state state;K_lEND}
  | "ensure"   {beg_state state;K_ENSURE}
  | "for"      {beg_state state;K_FOR}
  | "if"       {beg_state state;K_IF}
  | "in"       {beg_state state;K_IN}
  | "module"   {beg_state state;K_MODULE}
  | "not"      {beg_state state;K_NOT}
  | "or"       {beg_state state;K_OR}
  | "rescue"   {beg_state state;K_RESCUE}
  | "return"   {mid_state state;K_RETURN}
  | "then"     {beg_state state;K_THEN}
  | "unless"   {beg_state state;K_UNLESS}
  | "until"    {beg_state state;K_UNTIL}
  | "when"     {beg_state state;K_WHEN}
  | "while"    {beg_state state;K_WHILE}
  | "yield"    {mid_state state;K_YIELD}
  | "nil"      {mid_state state;K_NIL}
(* No longer lex separately
  | "self"     {mid_state state;K_SELF}
  | "super"    {mid_state state;K_SUPER}
  | "undef"    {beg_state state;K_UNDEF}
  | "break"    {beg_state state;K_BREAK}
  | "redo"     {beg_state state;K_REDO}
  | "true"     {mid_state state;K_TRUE}
  | "false"    {mid_state state;K_FALSE}
  | "retry"    {beg_state state;K_RETRY}
  | "next"     {beg_state state;K_NEXT}
      *)

  | "__END__" {end_lexbuf lexbuf}

  | ":" {atom state lexbuf}

  | " ::" { T_USCOPE }
  | "::" { choose_token state T_USCOPE T_SCOPE }
(*
  | (id_re? ("::" id_re)+) as s
      { mid_state state;let lst = Str.split_delim (Str.regexp "::") s in
	  T_SCOPED_ID lst
      }
*)
  | '$' id_re as id {mid_state state; T_GLOBAL_VAR id }
  | '@' {beg_state state; T_AT}
(*
  | '@' id_re as id {mid_state state; T_INST_VAR id }
  | "@@" id_re as id {mid_state state; T_CLASS_VAR id }
*)

  | id_re
      {mid_state state;match (Lexing.lexeme lexbuf).[0] with
	 | 'a'..'z' | '_' -> T_LID (Lexing.lexeme lexbuf)
	 | 'A'..'Z' -> T_UID (Lexing.lexeme lexbuf)
	 | _ -> failwith "unknown prefix char for ID: this shouldn't happen"
      }
(*  | '/'   {beg_state state;T_SLASH}*)
  | "/=" { let x = state.expr_state in
	    beg_state state;
	     match x with 
	       | Expr_Def
	       | Expr_Mid -> T_OP_ASGN (Lexing.lexeme lexbuf)
	       | Expr_Beg ->
		   let b = Buffer.create 32 in
		     Buffer.add_char b '=';
		     regexp_string state b lexbuf 
	 }

(*  | "/ " {T_SLASH}*)
  | '/'  (*{regexp_string state (Buffer.create 32) lexbuf}*)
      {let x = state.expr_state in
	 match x with 
	   | Expr_Def
	   | Expr_Mid -> beg_state state;T_SLASH
	   | Expr_Beg -> mid_state state;
	       regexp_string state (Buffer.create 32) lexbuf 
	}

  | ("$-" alphanum_re) as v {mid_state state; T_BUILTIN_VAR v }

  | ('$' ['0'-'9']+) as v {mid_state state;T_BUILTIN_VAR v}
  | ('$' [^'a'-'z''A'-'Z''#']) as v {mid_state state;T_BUILTIN_VAR v}

  | [' ''\t']+  {token state lexbuf}

and atom state = parse
  | ['+' '-' '*' '/' '!' '~' '<' '>' '=' '&' '|' '%']+ 
  | ('$'? '@'* id_re) '='?{mid_state state; T_ATOM (Lexing.lexeme lexbuf)}
  | ws_re
  | ""     {beg_state state; T_COLON}

and percent state = parse
  | (alpha_re* as modifier)(string_single_delim as d)
      {mid_state state; delim_string (Buffer.create 32) modifier d lexbuf}

  | (alpha_re* as modifier) '{'
      {mid_state state; delim_match_string (Buffer.create 32) modifier '{' '}' 1 lexbuf}
  | (alpha_re* as modifier) '<'
      {mid_state state; delim_match_string (Buffer.create 32) modifier '<' '>' 1 lexbuf}
  | (alpha_re* as modifier) '('
      {mid_state state; delim_match_string (Buffer.create 32) modifier '(' ')' 1 lexbuf}
  | (alpha_re* as modifier) '['
      {mid_state state; delim_match_string (Buffer.create 32) modifier '[' ']' 1 lexbuf}

  | "" {beg_state state;T_PERCENT}

and char_code = parse
  | [^'\n''\t''\\'] as c {T_FIXNUM (Char.code c)}
  | "\\s" {T_FIXNUM (Char.code ' ')}
  | "\\n" {T_FIXNUM (Char.code '\n')}
  | "\\t" {T_FIXNUM (Char.code '\t')}
  | "\\r" {T_FIXNUM (Char.code '\r')}

and comment = parse
  | [^'\n']* '\n' { incr_line lexbuf;T_EOL }

and delim_comment state = parse
  | "=end"{ token state lexbuf}
  | '\n' { incr_line lexbuf; delim_comment state lexbuf}
  | _ {delim_comment state lexbuf}

and single_string buf = parse
  | "#{" {interpret_string (single_string buf) lexbuf}
  | "\\\\" {Buffer.add_string buf "\\"; single_string buf lexbuf}
  | "\\'" {Buffer.add_char buf '\'';single_string buf lexbuf}
  | '\n' {incr_line lexbuf; Buffer.add_char buf '\n'; single_string buf lexbuf}
  |  [^'\''] as c {Buffer.add_char buf c; single_string buf lexbuf}
  | '\'' { T_SINGLE_STRING (Buffer.contents buf)}

and double_string buf = parse
  | "#{" {interpret_string (double_string buf) lexbuf}
  | "\\\\" {Buffer.add_string buf "\\"; double_string buf lexbuf}
  | "\\\"" {Buffer.add_char buf '"';double_string buf lexbuf}
  | '\n' {incr_line lexbuf; Buffer.add_char buf '\n'; double_string buf lexbuf}
  | [^'"'] as c {Buffer.add_char buf c; double_string buf lexbuf}
  | '"' {T_DOUBLE_STRING (Buffer.contents buf)}

and delim_string buf m delim = parse
  | "#{"   {interpret_string (delim_string buf m delim) lexbuf}
  | "\\" _ { Buffer.add_string buf (Lexing.lexeme lexbuf);
	     delim_string buf m delim lexbuf}
  | eof   {failwith "eof in delim?"}
  | '\n' {incr_line lexbuf; Buffer.add_char buf '\n'; 
	  delim_string buf m delim lexbuf}
  | _ as d { if d = delim then
	if m = "r"
	then regexp_modifier m (Buffer.contents buf) lexbuf
	else T_USER_STRING (m, Buffer.contents buf)
	     else (Buffer.add_char buf d;
		   delim_string buf m delim lexbuf)
	   }

and delim_match_string buf m d_start d_end level = parse
  | "#{"   {interpret_string (delim_match_string buf m d_start d_end level) lexbuf}
  | "\\" _ { Buffer.add_string buf (Lexing.lexeme lexbuf);
	     delim_match_string buf m d_start d_end level lexbuf}
  | eof   {failwith "eof in delim?"}
  | '\n' {incr_line lexbuf; Buffer.add_char buf '\n'; 
	  delim_match_string buf m d_start d_end level lexbuf}
  | _ as d
    { if d = d_end then begin
      if level = 1 then
	if m = "r"
	then regexp_modifier m (Buffer.contents buf) lexbuf
	else T_USER_STRING (m, Buffer.contents buf)
      else (Buffer.add_char buf d; 
	    delim_match_string buf m d_start d_end (level-1) lexbuf)
    end else 
	if d = d_start then
	  (Buffer.add_char buf d;
	   delim_match_string buf m d_start d_end (level+1) lexbuf)
      else (Buffer.add_char buf d; 
	    delim_match_string buf m d_start d_end level lexbuf)
}

and regexp_modifier m str = parse
  | (alpha_re* as modifiers) {assert (m = "r"); T_REGEXP (str,modifiers)}

(* FIXME *)
and regexp_string state buf = parse
  | "#{" {interpret_string (regexp_string state buf) lexbuf}
  | "\\\\" {Buffer.add_string buf "\\"; regexp_string state buf lexbuf}
  | "\\/" {Buffer.add_char buf '/'; regexp_string state buf lexbuf}
  | '/' (alpha_re* as modifiers) 
      {mid_state state; T_REGEXP (Buffer.contents buf, modifiers) } 
  | '\n' {incr_line lexbuf; Buffer.add_char buf '\n';
	  regexp_string state buf lexbuf}
  | eof   {failwith "eof in regexp?"}
  | _ as c {Buffer.add_char buf c; regexp_string state buf lexbuf}      

and heredoc_header state = parse
  | ['\'' '"']?(id_body+ as delim)['\'' '"']?
      {let id = fresh_heredoc state false delim in
	 mid_state state; T_HEREDOC id}
  | "" {beg_state state; T_LSHFT}

and heredoc state buf id delim = parse
  | ([^'\n']* as tok)
      { if tok = delim
	then begin
	  Hashtbl.add state.heredoc_tbl id (Buffer.contents buf);
	  T_EOL
	end else begin 
	  Buffer.add_string buf (Lexing.lexeme lexbuf);
	  heredoc state buf id delim lexbuf
	end
      }
  | '\n' {incr_line lexbuf; Buffer.add_char buf '\n'; 
	  heredoc state buf id delim lexbuf}
  | eof   {failwith "eof in heredoc?"}

and heredoc_lead state buf id delim = parse
  | [' ''\t']* ([^'\n']* as tok)
      {	if tok = delim
	then begin
	  Hashtbl.add state.heredoc_tbl id (Buffer.contents buf);
	  T_EOL
	end 
	else begin 
	  Buffer.add_string buf (Lexing.lexeme lexbuf);
	  heredoc_lead state buf id delim lexbuf
	end
      }
  | '\n' { incr_line lexbuf; Buffer.add_char buf '\n'; 
	   heredoc_lead state buf id delim lexbuf}
  | eof   {failwith "eof in heredoc?"}

and interpret_string cont = parse
  | [^'}']* '}' { cont lexbuf }

and end_lexbuf = parse
  | eof   {T_EOF}
  | _ {end_lexbuf lexbuf}
