
open Utils


type 'a default_list_ = [
   | `Param_Default of 'a * 'a default_list_
   | `Param_Star of 'a
   | `Param_Empty
  ]

type 'a param_typ_ = [
  | `Param_Var
  | `Param_t of 'a * 'a param_typ_
  | `Param_tuple of 'a * 'a * 'a param_typ_
  | 'a default_list_
  ]

module CG : ConstraintGraph.S

module type TypeKind_S = sig
  type t
  val format : Format.formatter -> t -> unit
  val format_small : Format.formatter -> t -> unit
  val unify : CG.t -> t -> t -> ctx:Log.ctx -> unit
end

module rec KTyp : 
sig
  include TypeKind_S
    
  val create : Log.ctx -> t
  val format : Format.formatter -> t -> unit
  val format_small : Format.formatter -> t -> unit

  val of_closed : KClass.t -> Log.ctx -> t
  val of_class : KClass.t -> Log.ctx -> t
  val of_tuple : KTuple.t -> Log.ctx -> t
  val of_union : t list -> Log.ctx -> t

  (** create a new open object containing a single method *)
  val of_method : KMethod.t -> Log.ctx -> t

  (** create a new open object containing a single field *)
  val of_field : KField.t -> Log.ctx -> t

  val of_param : KParam.t -> Log.ctx -> t

  val as_tuple : KTyp.t -> KTuple.t option
    (* return the class of t iff it is already a class *)
  val as_class : KTyp.t -> KClass.t option
    (* like as_class, but also unifies variables to classes *)
  val coerce_class : CG.t -> KTyp.t -> Log.ctx -> KClass.t option

  (** [new_instance cg ?name cls ctx] Creates a new 'instance' of the
      object 'cls' as would be done by Ruby's 'new' construct.  If the
      name argument is omitted, this function attempts to determine it
      from the argument 'cls'.  Typically, this argument can be
      ommitted when the 'cls' is a class (e.g. "String.new"), but may
      be supplied (e.g. "X=Object.new" requires X be supplied) *)
  val new_instance : CG.t -> ?name:string -> KTyp.t -> Log.ctx -> KTyp.t
    
  (** assert that 't' is actually a class with the given name *)
  val assert_named_class : CG.t -> t -> string -> Log.ctx -> unit
end
  
and KClass : 
sig
  include TypeKind_S

  val create : Log.ctx -> ?name:string -> ?inst:KClass.t -> 
    ?class_vars:KField.t Utils.StrMap.t Variable.t ->
    ?declared_subs:KClass.t list -> unit -> t

  val get_instance_class : KClass.t -> KClass.t

  val add_parent : CG.t -> parent:KTyp.t -> KClass.t -> ctx:Log.ctx -> unit
  val add_method : CG.t -> KMethod.t -> KClass.t -> Log.ctx -> unit
  val add_constant : CG.t -> string -> KTyp.t -> KClass.t -> Log.ctx -> unit

  val mem_constant : string -> KClass.t -> bool
  val lookup_constant : CG.t -> string -> KClass.t -> Log.ctx -> KTyp.t
  val lookup_inst_var : CG.t -> string -> KClass.t -> Log.ctx -> KTyp.t
  val lookup_class_var : CG.t -> string -> KClass.t -> Log.ctx -> KTyp.t

  val alias_method : CG.t -> exists:string -> link:string -> KClass.t 
    -> KTyp.t -> Log.ctx -> unit

  val module_function : CG.t -> string -> KClass.t -> Log.ctx -> unit
end


and KMethod : 
sig 
  include TypeKind_S
  val create : Log.ctx -> string -> ?func:KFunc.t -> ?block:KBlock.t -> unit -> KMethod.t
  val method_name : t -> string
  (* Produces a monomorphic instance of the method [t].
     Intersection types are not yet supported, so in that case, we
     return None. *)
  val instantiate_annotation : CG.t -> t -> Log.ctx -> t option
end

and KParam : 
sig
  type params = KTyp.t param_typ_
  type t = params Variable.t
  val create : Log.ctx -> KParam.t
  val format : Format.formatter -> t -> unit
  val format_small : Format.formatter -> t -> unit
end

and KFunc : 
sig
  include TypeKind_S
  val create : Log.ctx -> ?self:KTyp.t -> ?args:KParam.t -> ?ret:KTyp.t -> unit -> KFunc.t
  val get_self : t -> KTyp.t
  val get_params : t -> KParam.t
  val get_ret : t -> KTyp.t
end

and KBlock : 
sig
  include TypeKind_S
  val create : Log.ctx -> ?bsig:KFunc.t -> unit -> KBlock.t
  val no_block : Log.ctx -> KBlock.t
  val to_func : CG.t -> KBlock.t -> Log.ctx -> KFunc.t option
end

and KTuple : 
sig
  include TypeKind_S
  val create : Log.ctx -> KTyp.t list -> elt:KTyp.t -> array:KTyp.t -> KTuple.t
  val unfold : t -> KTyp.t list
end

and KDict   : TypeKind_S

and KField  : sig 
  include TypeKind_S
  val create : Log.ctx -> string -> ?t:KTyp.t -> unit -> KField.t
end

module type Constraint_Mod = sig
  module ConKind : TypeKind_S
  val sub_constraint : CG.t -> ConKind.t -> ConKind.t -> Log.ctx -> unit
end

module ConTyp : Constraint_Mod with module ConKind = KTyp
module ConMethod : Constraint_Mod with module ConKind = KMethod
module ConSuper : Constraint_Mod with module ConKind = KTyp

module Deprecated : sig
  val get_class : CG.t -> KTyp.t -> Log.ctx -> KClass.t
  val fresh_instance_with_params : CG.t -> KTyp.t -> KTyp.t list -> Log.ctx -> KTyp.t
end

type any_list = [
   | `Param_Default of KTyp.t * any_list
   | `Param_Star of KTyp.t
   | `Param_Empty
   | `Param_Var
   | `Param_t of KTyp.t * any_list
   | `Param_tuple of KTyp.t * KTyp.t * any_list
]
 
module Annotations : sig

  type annot_env
    
  val empty_annot_env : KClass.t -> annot_env
    
  val annot_set_cur_class : annot_env -> KClass.t -> annot_env
    
  val class_of_annotation : CG.t -> annot_env -> Annotation.t
    -> Log.ctx -> Log.pos -> annot_env * KTyp.t
      
  val mt_of_annotation : CG.t -> annot_env -> Annotation.t
    -> Log.ctx -> Log.pos -> annot_env * KMethod.t

  val t_expr_of_annotation : CG.t -> annot_env -> Annotation.t
    -> Log.ctx -> Log.pos -> KTyp.t
end
